#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jan 14 15:39:29 2020

@author: ljia
"""
import multiprocessing
import functools
from gklearn.preimage.utils import generate_median_preimages_by_class

		
def test_median_preimage_generator():
	"""MAO, Treelet, using CONSTANT, symbolic only.
	"""
	from gklearn.utils.kernels import polynomialkernel
	# set parameters.
	ds_name = 'MAO' #
	mpg_options = {'fit_method': 'k-graphs',
				   'init_ecc': [4, 4, 2, 1, 1, 1], #
				   'ds_name': ds_name,
				   'parallel': True, # False
				   'time_limit_in_sec': 0,
				   'max_itrs': 3, # 
				   'max_itrs_without_update': 3,
				   'epsilon_residual': 0.01,
				   'epsilon_ec': 0.1,
				   'verbose': 2}
	pkernel = functools.partial(polynomialkernel, d=4, c=1e+7)
	kernel_options = {'name': 'Treelet', #
				      'sub_kernel': pkernel,
 					  'parallel': 'imap_unordered', 
                      # 'parallel': None, 
					  'n_jobs': multiprocessing.cpu_count(),
					  'normalize': True,
					  'verbose': 2}
	ged_options = {'method': 'IPFP',
				   'initialization_method': 'RANDOM', # 'NODE'
				   'initial_solutions': 1, # 1
				   'edit_cost': 'CONSTANT', # 
				   'attr_distance': 'euclidean',
				   'ratio_runs_from_initial_solutions': 1,
				   'threads': multiprocessing.cpu_count(),
				   'init_option': 'EAGER_WITHOUT_SHUFFLED_COPIES'}
	mge_options = {'init_type': 'MEDOID',
				   'random_inits': 10,
				   'time_limit': 600,
				   'verbose': 2,
				   'refine': False}
	save_results = True
	dir_save = ds_name + '.' + kernel_options['name'] + '.symb.pytest/'
	irrelevant_labels = {'node_attrs': ['x', 'y', 'z'], 'edge_labels': ['bond_stereo']} # 
	edge_required = False #
	
	# print settings.
	print('parameters:')
	print('dataset name:', ds_name)
	print('mpg_options:', mpg_options)
	print('kernel_options:', kernel_options)
	print('ged_options:', ged_options)
	print('mge_options:', mge_options)
	print('save_results:', save_results)
	print('irrelevant_labels:', irrelevant_labels)
	print()
	
	# generate preimages.
	for fit_method in ['k-graphs', 'expert', 'random']:
		print('\n-------------------------------------')
		print('fit method:', fit_method, '\n')
		mpg_options['fit_method'] = fit_method
		try:
			generate_median_preimages_by_class(ds_name, mpg_options, kernel_options, ged_options, mge_options, save_results=save_results, save_medians=True, plot_medians=True, load_gm='auto', dir_save=dir_save, irrelevant_labels=irrelevant_labels, edge_required=edge_required, cut_range=range(0, 4))
		except Exception as exception:
			assert False, exception
			
	
if __name__ == '__main__':
	test_median_preimage_generator()


# -*- coding: utf-8 -*-
"""model_selection_old.ipynb

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1uVkl7scNgEPrimX8ks6iEC5ijuhB8L_D

**This script demonstrates how to compute a graph kernel.**
---

**0.   Install `graphkit-learn`.**
"""

"""**1. Perform model seletion and classification.**"""

from gklearn.utils import model_selection_for_precomputed_kernel
from gklearn.kernels import untilhpathkernel
import numpy as np

# Set parameters.
datafile = '../../../datasets/MUTAG/MUTAG_A.txt'
param_grid_precomputed = {'depth': np.linspace(1, 10, 10),
                          'k_func': ['MinMax', 'tanimoto'],
                          'compute_method': ['trie']}
param_grid = {'C': np.logspace(-10, 10, num=41, base=10)}

# Perform model selection and classification.
model_selection_for_precomputed_kernel(
	datafile, # The path of dataset file.
	untilhpathkernel, # The graph kernel used for estimation.
	param_grid_precomputed, # The parameters used to compute gram matrices.
	param_grid, # The penelty Parameters used for penelty items.
	'classification', # Or 'regression'.
	NUM_TRIALS=30, # The number of the random trials of the outer CV loop.
	ds_name='MUTAG', # The name of the dataset.
	n_jobs=1,
	verbose=True)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 17 17:52:23 2020

@author: ljia
"""
from gklearn.ged.edit_costs import EditCost


class Constant(EditCost):
	"""Implements constant edit cost functions.
	"""

	
	def __init__(self, node_ins_cost=1, node_del_cost=1, node_rel_cost=1, edge_ins_cost=1, edge_del_cost=1, edge_rel_cost=1):
		self._node_ins_cost = node_ins_cost
		self._node_del_cost = node_del_cost
		self._node_rel_cost = node_rel_cost
		self._edge_ins_cost = edge_ins_cost
		self._edge_del_cost = edge_del_cost
		self._edge_rel_cost = edge_rel_cost
		
		
	def node_ins_cost_fun(self, node_label):
		return self._node_ins_cost
	
	
	def node_del_cost_fun(self, node_label):
		return self._node_del_cost
	
	
	def node_rel_cost_fun(self, node_label_1, node_label_2):
		if node_label_1 != node_label_2:
			return self._node_rel_cost
		return 0
	
	
	def edge_ins_cost_fun(self, edge_label):
		return self._edge_ins_cost
	
	
	def edge_del_cost_fun(self, edge_label):
		return self._edge_del_cost
	
	
	def edge_rel_cost_fun(self, edge_label_1, edge_label_2):
		if edge_label_1 != edge_label_2:
			return self._edge_rel_cost
		return 0


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 20 14:25:49 2020

@author: ljia
"""

DATABASES = {
	'greyc': 'https://brunl01.users.greyc.fr/CHEMISTRY/',
	'iam': 'https://iapr-tc15.greyc.fr/IAM/',
	'tudataset': 'http://graphlearning.io/docs/datasets/',
}

### -------- database greyc -------- ###
GREYC_META = {
	'ACE': {
		'database': 'greyc',
		'reference': None,
		'dataset_size': 32,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 52,
		'ave_edge_num': None,
		'node_labeled': None,
		'edge_labeled': None,
		'node_attr_dim': None,
		'geometry': None,
		'edge_attr_dim': None,
		'url': 'https://brunl01.users.greyc.fr/CHEMISTRY/ACEDataset.tar',
		'domain': 'small molecules',
		'train_valid_test': [],
		'stereoisomerism': True,
		'load_files': ['data.ds'],
		'extra_params': {'gformat': 'cml'}
	},
	'Acyclic': {
		'database': 'greyc',
		'reference': None,
		'dataset_size': 183,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 8.15,
		'ave_edge_num': 7.15,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://brunl01.users.greyc.fr/CHEMISTRY/Acyclic.tar.gz',
		'domain': 'small molecules',
		'train_valid_test': None,
		'stereoisomerism': False,
		'load_files': ['dataset_bps.ds'],
	},
	'AIDS_greyc': {
		'database': 'greyc',
		'reference': None,
		'dataset_size': 2000,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 15.69,
		'ave_edge_num': 16.20,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': None,
		'geometry': None,
		'edge_attr_dim': None,
		'url': 'https://iapr-tc15.greyc.fr/IAM/AIDS.zip',
		'domain': 'small molecules',
		'train_valid_test': ['data/train.cxl', 'data/valid.cxl',
		                     'data/test.cxl', ],
		'stereoisomerism': False,
		'load_files': ['data/'],
	},
	'Alkane': {
		'database': 'greyc',
		'reference': None,
		'dataset_size': 150,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 8.87,
		'ave_edge_num': 7.87,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://brunl01.users.greyc.fr/CHEMISTRY/alkane_dataset.tar.gz',
		'domain': 'small molecules',
		'train_valid_test': None,
		'stereoisomerism': False,
		'load_files': ['dataset.ds', 'dataset_boiling_point_names.txt'],
	},
	'Chiral': {
		'database': 'greyc',
		'reference': None,
		'dataset_size': 35,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 21.29,
		'ave_edge_num': None,
		'node_labeled': None,
		'edge_labeled': None,
		'node_attr_dim': None,
		'geometry': None,
		'edge_attr_dim': None,
		'url': 'https://brunl01.users.greyc.fr/CHEMISTRY/DatasetAcyclicChiral.tar',
		'domain': 'small molecules',
		'train_valid_test': [],
		'stereoisomerism': True,
		'load_files': ['data.txt'],
		'extra_params': {'gformat': 'cml'}
	},
	'MAO': {
		'database': 'greyc',
		'reference': None,
		'dataset_size': 68,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 18.38,
		'ave_edge_num': 19.63,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://brunl01.users.greyc.fr/CHEMISTRY/mao.tgz',
		'domain': 'small molecules',
		'train_valid_test': None,
		'stereoisomerism': False,
		'load_files': ['dataset.ds'],
	},
	'Monoterpenoides': {
		'database': 'greyc',
		'reference': None,
		'dataset_size': 382,
		'class_number': 10,
		'task_type': 'classification',
		'ave_node_num': 10,
		'ave_edge_num': None,
		'node_labeled': None,
		'edge_labeled': None,
		'node_attr_dim': None,
		'geometry': None,
		'edge_attr_dim': None,
		'url': 'https://brunl01.users.greyc.fr/CHEMISTRY/monoterpenoides.tar.gz',
		'domain': 'small molecules',
		'train_valid_test': None,
		'stereoisomerism': False,
		'load_files': ['dataset_10+.ds'],
	},
	'PAH': {
		'database': 'greyc',
		'reference': None,
		'dataset_size': 94,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 20.7,
		'ave_edge_num': 24.43,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://brunl01.users.greyc.fr/CHEMISTRY/PAH.tar.gz',
		'domain': 'small molecules',
		'train_valid_test': ['trainset_0.ds', None, 'testset_0.ds'],
		'stereoisomerism': False,
		'load_files': [['trainset_0.ds', 'testset_0.ds']],
	},
	'PTC': {
		'database': 'greyc',
		'reference': None,
		'dataset_size': 416,
		'class_number': None,
		'task_type': 'classification',
		'ave_node_num': 14.4,
		'ave_edge_num': None,
		'node_labeled': None,
		'edge_labeled': None,
		'node_attr_dim': None,
		'geometry': None,
		'edge_attr_dim': None,
		'url': 'https://brunl01.users.greyc.fr/CHEMISTRY/ptc.tgz',
		'domain': 'small molecules',
		'train_valid_test': None,
		'stereoisomerism': False,
		'load_files': [],
		'extra_info': 'This dataset has test and train datasets. Select gender between mm, fm, mr, fr. \ndataloader = DataLoader(\'Ptc\',root = ..., option = \'mm\') \ntest,train = dataloader.dataset \nGs_test, y_test = test \nGs_train_, y_train = train',
	},
	'Steroid': {
		'database': 'greyc',
		'reference': None,
		'dataset_size': 64,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 75.11,
		'ave_edge_num': None,
		'node_labeled': None,
		'edge_labeled': None,
		'node_attr_dim': None,
		'geometry': None,
		'edge_attr_dim': None,
		'url': 'https://brunl01.users.greyc.fr/CHEMISTRY/SteroidDataset.tar',
		'domain': 'small molecules',
		'train_valid_test': ['trainset_0.ds', None, 'testset_0.ds'],
		'stereoisomerism': False,
		'load_files': ['dataWithOutsider.ds'],
		'extra_params': {'gformat': 'cml'}
	},
	'Vitamin_D': {
		'database': 'greyc',
		'reference': None,
		'dataset_size': 69,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 76.91,
		'ave_edge_num': None,
		'node_labeled': None,
		'edge_labeled': None,
		'node_attr_dim': None,
		'geometry': None,
		'edge_attr_dim': None,
		'url': 'https://brunl01.users.greyc.fr/CHEMISTRY/DatasetVitamin.tar',
		'domain': 'small molecules',
		'train_valid_test': [],
		'stereoisomerism': True,
		'load_files': ['data.txt'],
		'extra_params': {'gformat': 'cml'}
	},
}

### -------- database iam -------- ###
# @todo: several datasets in this database are included in TUDataset. However they do not include train/valid/test sets.
IAM_META = {
	'GREC': {
		'database': 'iam',
		'reference': None,
		'dataset_size': None,
		'class_number': None,
		'task_type': None,
		'ave_node_num': None,
		'ave_edge_num': None,
		'node_labeled': None,
		'edge_labeled': None,
		'node_attr_dim': None,
		'geometry': None,
		'edge_attr_dim': None,
		'url': 'https://iapr-tc15.greyc.fr/IAM/GREC.zip',
		'domain': None,
		'train_valid_test': ['data/test.cxl', 'data/train.cxl',
		                     'data/valid.cxl'],
		'load_files': [['data/test.cxl', 'data/train.cxl', 'data/valid.cxl']],
	},
	'Web': {
		'database': 'iam',
		'reference': None,
		'dataset_size': None,
		'class_number': None,
		'task_type': None,
		'ave_node_num': None,
		'ave_edge_num': None,
		'node_labeled': None,
		'edge_labeled': None,
		'node_attr_dim': None,
		'geometry': None,
		'edge_attr_dim': None,
		'url': 'https://iapr-tc15.greyc.fr/IAM/Web.zip',
		'domain': None,
		'train_valid_test': ['data/test.cxl', 'data/train.cxl',
		                     'data/valid.cxl'],
		'load_files': [['data/test.cxl', 'data/train.cxl', 'data/valid.cxl']],
	},
}

### -------- database tudataset -------- ###
TUDataset_META = {
	### small molecules
	'AIDS': {
		'database': 'tudataset',
		'reference': '[16,17]',
		'dataset_size': 2000,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 15.69,
		'ave_edge_num': 16.2,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 4,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/AIDS.zip',
		'domain': 'small molecules',
	},
	'alchemy_full': {
		'database': 'tudataset',
		'reference': '[29]',
		'dataset_size': 202579,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 10.1,
		'ave_edge_num': 10.44,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 3,
		'geometry': '3D, RI',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/alchemy_full.zip',
		'domain': 'small molecules',
	},
	'aspirin': {
		'database': 'tudataset',
		'reference': '[36]',
		'dataset_size': 111763,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 21.0,
		'ave_edge_num': 151.52,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 6,
		'geometry': '3D, RI',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/aspirin.zip',
		'domain': 'small molecules',
	},
	'benzene': {
		'database': 'tudataset',
		'reference': '[36]',
		'dataset_size': 527984,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 12.0,
		'ave_edge_num': 64.94,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 6,
		'geometry': '3D, RI',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/benzene.zip',
		'domain': 'small molecules',
	},
	'BZR': {
		'database': 'tudataset',
		'reference': '[7]',
		'dataset_size': 405,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 35.75,
		'ave_edge_num': 38.36,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 3,
		'geometry': '3D, RI',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/BZR.zip',
		'domain': 'small molecules',
	},
	'BZR_MD': {
		'database': 'tudataset',
		'reference': '[7,23]',
		'dataset_size': 306,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 21.3,
		'ave_edge_num': 225.06,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 1,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/BZR_MD.zip',
		'domain': 'small molecules',
	},
	'COX2': {
		'database': 'tudataset',
		'reference': '[7]',
		'dataset_size': 467,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 41.22,
		'ave_edge_num': 43.45,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 3,
		'geometry': '3D, RI',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/COX2.zip',
		'domain': 'small molecules',
	},
	'COX2_MD': {
		'database': 'tudataset',
		'reference': '[7,23]',
		'dataset_size': 303,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 26.28,
		'ave_edge_num': 335.12,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 1,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/COX2_MD.zip',
		'domain': 'small molecules',
	},
	'DHFR': {
		'database': 'tudataset',
		'reference': '[7]',
		'dataset_size': 467,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 42.43,
		'ave_edge_num': 44.54,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 3,
		'geometry': '3D, RI',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/DHFR.zip',
		'domain': 'small molecules',
	},
	'DHFR_MD': {
		'database': 'tudataset',
		'reference': '[7,23]',
		'dataset_size': 393,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 23.87,
		'ave_edge_num': 283.01,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 1,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/DHFR_MD.zip',
		'domain': 'small molecules',
	},
	'ER_MD': {
		'database': 'tudataset',
		'reference': '[7,23]',
		'dataset_size': 446,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 21.33,
		'ave_edge_num': 234.85,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 1,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/ER_MD.zip',
		'domain': 'small molecules',
	},
	'ethanol': {
		'database': 'tudataset',
		'reference': '[36]',
		'dataset_size': 455093,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 9.0,
		'ave_edge_num': 36.0,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 6,
		'geometry': '3D, RI',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/ethanol.zip',
		'domain': 'small molecules',
	},
	'FRANKENSTEIN': {
		'database': 'tudataset',
		'reference': '[15]',
		'dataset_size': 4337,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 16.9,
		'ave_edge_num': 17.88,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 780,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/FRANKENSTEIN.zip',
		'domain': 'small molecules',
	},
	'malonaldehyde': {
		'database': 'tudataset',
		'reference': '[36]',
		'dataset_size': 893238,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 9.0,
		'ave_edge_num': 36.0,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 6,
		'geometry': '3D, RI',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/malonaldehyde.zip',
		'domain': 'small molecules',
	},
	'MCF-7': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 27770,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 26.39,
		'ave_edge_num': 28.52,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/MCF-7.zip',
		'domain': 'small molecules',
	},
	'MCF-7H': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 27770,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 47.3,
		'ave_edge_num': 49.43,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/MCF-7H.zip',
		'domain': 'small molecules',
	},
	'MOLT-4': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 39765,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 26.09,
		'ave_edge_num': 28.13,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/MOLT-4.zip',
		'domain': 'small molecules',
	},
	'MOLT-4H': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 39765,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 46.7,
		'ave_edge_num': 48.73,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/MOLT-4H.zip',
		'domain': 'small molecules',
	},
	'Mutagenicity': {
		'database': 'tudataset',
		'reference': '[16,20]',
		'dataset_size': 4337,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 30.32,
		'ave_edge_num': 30.77,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Mutagenicity.zip',
		'domain': 'small molecules',
	},
	'MUTAG': {
		'database': 'tudataset',
		'reference': '[1,23]',
		'dataset_size': 188,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.93,
		'ave_edge_num': 19.79,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/MUTAG.zip',
		'domain': 'small molecules',
	},
	'naphthalene': {
		'database': 'tudataset',
		'reference': '[36]',
		'dataset_size': 226256,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 18.0,
		'ave_edge_num': 127.37,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 6,
		'geometry': '3D, RI',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/naphthalene.zip',
		'domain': 'small molecules',
	},
	'NCI1': {
		'database': 'tudataset',
		'reference': '[8,9,22]',
		'dataset_size': 4110,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 29.87,
		'ave_edge_num': 32.3,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/NCI1.zip',
		'domain': 'small molecules',
	},
	'NCI109': {
		'database': 'tudataset',
		'reference': '[8,9,22]',
		'dataset_size': 4127,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 29.68,
		'ave_edge_num': 32.13,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/NCI109.zip',
		'domain': 'small molecules',
	},
	'NCI-H23': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 40353,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 26.07,
		'ave_edge_num': 28.1,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/NCI-H23.zip',
		'domain': 'small molecules',
	},
	'NCI-H23H': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 40353,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 46.67,
		'ave_edge_num': 48.69,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/NCI-H23H.zip',
		'domain': 'small molecules',
	},
	'OVCAR-8': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 40516,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 26.07,
		'ave_edge_num': 28.1,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/OVCAR-8.zip',
		'domain': 'small molecules',
	},
	'OVCAR-8H': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 40516,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 46.67,
		'ave_edge_num': 48.7,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/OVCAR-8H.zip',
		'domain': 'small molecules',
	},
	'P388': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 41472,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 22.11,
		'ave_edge_num': 23.55,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/P388.zip',
		'domain': 'small molecules',
	},
	'P388H': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 41472,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 40.44,
		'ave_edge_num': 41.88,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/P388H.zip',
		'domain': 'small molecules',
	},
	'PC-3': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 27509,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 26.35,
		'ave_edge_num': 28.49,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/PC-3.zip',
		'domain': 'small molecules',
	},
	'PC-3H': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 27509,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 47.19,
		'ave_edge_num': 49.32,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/PC-3H.zip',
		'domain': 'small molecules',
	},
	'PTC_FM': {
		'database': 'tudataset',
		'reference': '[2,23]',
		'dataset_size': 349,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 14.11,
		'ave_edge_num': 14.48,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/PTC_FM.zip',
		'domain': 'small molecules',
	},
	'PTC_FR': {
		'database': 'tudataset',
		'reference': '[2,23]',
		'dataset_size': 351,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 14.56,
		'ave_edge_num': 15.0,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/PTC_FR.zip',
		'domain': 'small molecules',
	},
	'PTC_MM': {
		'database': 'tudataset',
		'reference': '[2,23]',
		'dataset_size': 336,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 13.97,
		'ave_edge_num': 14.32,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/PTC_MM.zip',
		'domain': 'small molecules',
	},
	'PTC_MR': {
		'database': 'tudataset',
		'reference': '[2,23]',
		'dataset_size': 344,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 14.29,
		'ave_edge_num': 14.69,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/PTC_MR.zip',
		'domain': 'small molecules',
	},
	'QM9': {
		'database': 'tudataset',
		'reference': '[33,34,35]',
		'dataset_size': 129433,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 18.03,
		'ave_edge_num': 18.63,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 16,
		'geometry': '3D, RI',
		'edge_attr_dim': 4,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/QM9.zip',
		'domain': 'small molecules',
	},
	'salicylic_acid': {
		'database': 'tudataset',
		'reference': '[36]',
		'dataset_size': 220232,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 16.0,
		'ave_edge_num': 104.13,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 6,
		'geometry': '3D, RI',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/salicylic_acid.zip',
		'domain': 'small molecules',
	},
	'SF-295': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 40271,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 26.06,
		'ave_edge_num': 28.08,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/SF-295.zip',
		'domain': 'small molecules',
	},
	'SF-295H': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 40271,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 46.65,
		'ave_edge_num': 48.68,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/SF-295H.zip',
		'domain': 'small molecules',
	},
	'SN12C': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 40004,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 26.08,
		'ave_edge_num': 28.11,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/SN12C.zip',
		'domain': 'small molecules',
	},
	'SN12CH': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 40004,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 46.69,
		'ave_edge_num': 48.71,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/SN12CH.zip',
		'domain': 'small molecules',
	},
	'SW-620': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 40532,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 26.05,
		'ave_edge_num': 28.08,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/SW-620.zip',
		'domain': 'small molecules',
	},
	'SW-620H': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 40532,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 46.62,
		'ave_edge_num': 48.65,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/SW-620H.zip',
		'domain': 'small molecules',
	},
	'toluene': {
		'database': 'tudataset',
		'reference': '[36]',
		'dataset_size': 342791,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 15.0,
		'ave_edge_num': 96.15,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 6,
		'geometry': '3D, RI',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/toluene.zip',
		'domain': 'small molecules',
	},
	'Tox21_AhR_training': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 8169,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 18.09,
		'ave_edge_num': 18.5,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_AhR_training.zip',
		'domain': 'small molecules',
	},
	'Tox21_AhR_testing': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 272,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 22.13,
		'ave_edge_num': 23.05,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_AhR_testing.zip',
		'domain': 'small molecules',
	},
	'Tox21_AhR_evaluation': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 607,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.64,
		'ave_edge_num': 18.06,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_AhR_evaluation.zip',
		'domain': 'small molecules',
	},
	'Tox21_AR_training': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 9362,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 18.39,
		'ave_edge_num': 18.84,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_AR_training.zip',
		'domain': 'small molecules',
	},
	'Tox21_AR_testing': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 292,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 22.35,
		'ave_edge_num': 23.32,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_AR_testing.zip',
		'domain': 'small molecules',
	},
	'Tox21_AR_evaluation': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 585,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.99,
		'ave_edge_num': 18.45,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_AR_evaluation.zip',
		'domain': 'small molecules',
	},
	'Tox21_AR-LBD_training': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 8599,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.77,
		'ave_edge_num': 18.16,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_AR-LBD_training.zip',
		'domain': 'small molecules',
	},
	'Tox21_AR-LBD_testing': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 253,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 21.85,
		'ave_edge_num': 22.73,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_AR-LBD_testing.zip',
		'domain': 'small molecules',
	},
	'Tox21_AR-LBD_evaluation': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 580,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.09,
		'ave_edge_num': 17.42,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_AR-LBD_evaluation.zip',
		'domain': 'small molecules',
	},
	'Tox21_ARE_training': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 7167,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 16.28,
		'ave_edge_num': 16.52,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_ARE_training.zip',
		'domain': 'small molecules',
	},
	'Tox21_ARE_testing': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 234,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 21.99,
		'ave_edge_num': 22.91,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_ARE_testing.zip',
		'domain': 'small molecules',
	},
	'Tox21_ARE_evaluation': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 552,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.01,
		'ave_edge_num': 17.33,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_ARE_evaluation.zip',
		'domain': 'small molecules',
	},
	'Tox21_aromatase_training': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 7226,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.5,
		'ave_edge_num': 17.79,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_aromatase_training.zip',
		'domain': 'small molecules',
	},
	'Tox21_aromatase_testing': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 214,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 21.65,
		'ave_edge_num': 22.36,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_aromatase_testing.zip',
		'domain': 'small molecules',
	},
	'Tox21_aromatase_evaluation': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 528,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 16.74,
		'ave_edge_num': 16.99,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_aromatase_evaluation.zip',
		'domain': 'small molecules',
	},
	'Tox21_ATAD5_training': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 9091,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.89,
		'ave_edge_num': 18.3,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_ATAD5_training.zip',
		'domain': 'small molecules',
	},
	'Tox21_ATAD5_testing': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 272,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 21.99,
		'ave_edge_num': 22.89,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_ATAD5_testing.zip',
		'domain': 'small molecules',
	},
	'Tox21_ATAD5_evaluation': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 619,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.68,
		'ave_edge_num': 18.11,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_ATAD5_evaluation.zip',
		'domain': 'small molecules',
	},
	'Tox21_ER_training': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 7697,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.58,
		'ave_edge_num': 17.94,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_ER_training.zip',
		'domain': 'small molecules',
	},
	'Tox21_ER_testing': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 265,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 22.16,
		'ave_edge_num': 23.13,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_ER_testing.zip',
		'domain': 'small molecules',
	},
	'Tox21_ER_evaluation': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 515,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.66,
		'ave_edge_num': 18.1,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_ER_evaluation.zip',
		'domain': 'small molecules',
	},
	'Tox21_ER-LBD_training': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 8753,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 18.06,
		'ave_edge_num': 18.47,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_ER-LBD_training.zip',
		'domain': 'small molecules',
	},
	'Tox21_ER-LBD_testing': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 287,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 22.28,
		'ave_edge_num': 23.23,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_ER-LBD_testing.zip',
		'domain': 'small molecules',
	},
	'Tox21_ER-LBD_evaluation': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 599,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.75,
		'ave_edge_num': 18.17,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_ER-LBD_evaluation.zip',
		'domain': 'small molecules',
	},
	'Tox21_HSE_training': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 8150,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 16.72,
		'ave_edge_num': 17.04,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_HSE_training.zip',
		'domain': 'small molecules',
	},
	'Tox21_HSE_testing': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 267,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 22.07,
		'ave_edge_num': 23.0,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_HSE_testing.zip',
		'domain': 'small molecules',
	},
	'Tox21_HSE_evaluation': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 607,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.61,
		'ave_edge_num': 18.01,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_HSE_evaluation.zip',
		'domain': 'small molecules',
	},
	'Tox21_MMP_training': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 7320,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.49,
		'ave_edge_num': 17.83,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_MMP_training.zip',
		'domain': 'small molecules',
	},
	'Tox21_MMP_testing': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 238,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 21.68,
		'ave_edge_num': 22.55,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_MMP_testing.zip',
		'domain': 'small molecules',
	},
	'Tox21_MMP_evaluation': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 541,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 16.67,
		'ave_edge_num': 16.88,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_MMP_evaluation.zip',
		'domain': 'small molecules',
	},
	'Tox21_p53_training': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 8634,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.79,
		'ave_edge_num': 18.19,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_p53_training.zip',
		'domain': 'small molecules',
	},
	'Tox21_p53_testing': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 269,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 22.14,
		'ave_edge_num': 23.04,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_p53_testing.zip',
		'domain': 'small molecules',
	},
	'Tox21_p53_evaluation': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 613,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.34,
		'ave_edge_num': 17.72,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_p53_evaluation.zip',
		'domain': 'small molecules',
	},
	'Tox21_PPAR-gamma_training': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 8184,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.23,
		'ave_edge_num': 17.55,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_PPAR-gamma_training.zip',
		'domain': 'small molecules',
	},
	'Tox21_PPAR-gamma_testing': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 267,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 22.04,
		'ave_edge_num': 22.93,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_PPAR-gamma_testing.zip',
		'domain': 'small molecules',
	},
	'Tox21_PPAR-gamma_evaluation': {
		'database': 'tudataset',
		'reference': '[24]',
		'dataset_size': 602,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 17.38,
		'ave_edge_num': 17.77,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Tox21_PPAR-gamma_evaluation.zip',
		'domain': 'small molecules',
	},
	'UACC257': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 39988,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 26.09,
		'ave_edge_num': 28.12,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/UACC257.zip',
		'domain': 'small molecules',
	},
	'UACC257H': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 39988,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 46.68,
		'ave_edge_num': 48.71,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/UACC257H.zip',
		'domain': 'small molecules',
	},
	'uracil': {
		'database': 'tudataset',
		'reference': '[36]',
		'dataset_size': 133770,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 12.0,
		'ave_edge_num': 64.44,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 6,
		'geometry': '3D, RI',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/uracil.zip',
		'domain': 'small molecules',
	},
	'Yeast': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 79601,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 21.54,
		'ave_edge_num': 22.84,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Yeast.zip',
		'domain': 'small molecules',
	},
	'YeastH': {
		'database': 'tudataset',
		'reference': '[28]',
		'dataset_size': 79601,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 39.44,
		'ave_edge_num': 40.74,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/YeastH.zip',
		'domain': 'small molecules',
	},
	'ZINC_full': {
		'database': 'tudataset',
		'reference': '[31]',
		'dataset_size': 249456,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 23.14,
		'ave_edge_num': 24.91,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/ZINC_full.zip',
		'domain': 'small molecules',
	},
	'ZINC_test': {
		'database': 'tudataset',
		'reference': '[31]',
		'dataset_size': 5000,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 23.1,
		'ave_edge_num': 24.83,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/ZINC_test.zip',
		'domain': 'small molecules',
	},
	'ZINC_train': {
		'database': 'tudataset',
		'reference': '[31]',
		'dataset_size': 220011,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 23.15,
		'ave_edge_num': 24.91,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/ZINC_train.zip',
		'domain': 'small molecules',
	},
	'ZINC_val': {
		'database': 'tudataset',
		'reference': '[31]',
		'dataset_size': 24445,
		'class_number': None,
		'task_type': 'regression',
		'ave_node_num': 23.13,
		'ave_edge_num': 24.88,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/ZINC_val.zip',
		'domain': 'small molecules',
	},

	### bioinformatics
	'DD': {
		'database': 'tudataset',
		'reference': '[6,22]',
		'dataset_size': 1178,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 284.32,
		'ave_edge_num': 715.66,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/DD.zip',
		'domain': 'bioinformatics',
	},
	'ENZYMES': {
		'database': 'tudataset',
		'reference': '[4,5]',
		'dataset_size': 600,
		'class_number': 6,
		'task_type': 'classification',
		'ave_node_num': 32.63,
		'ave_edge_num': 62.14,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 18,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'http://www.chrsmrrs.com/graphkerneldatasets/ENZYMES.zip',
		'domain': 'bioinformatics',
	},
	'KKI': {
		'database': 'tudataset',
		'reference': '[26]',
		'dataset_size': 83,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 26.96,
		'ave_edge_num': 48.42,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/KKI.zip',
		'domain': 'bioinformatics',
	},
	'OHSU': {
		'database': 'tudataset',
		'reference': '[26]',
		'dataset_size': 79,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 82.01,
		'ave_edge_num': 199.66,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/OHSU.zip',
		'domain': 'bioinformatics',
	},
	'Peking_1': {
		'database': 'tudataset',
		'reference': '[26]',
		'dataset_size': 85,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 39.31,
		'ave_edge_num': 77.35,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Peking_1.zip',
		'domain': 'bioinformatics',
	},
	'PROTEINS': {
		'database': 'tudataset',
		'reference': '[4,6]',
		'dataset_size': 1113,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 39.06,
		'ave_edge_num': 72.82,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 1,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/PROTEINS.zip',
		'domain': 'bioinformatics',
	},
	'PROTEINS_full': {
		'database': 'tudataset',
		'reference': '[4,6]',
		'dataset_size': 1113,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 39.06,
		'ave_edge_num': 72.82,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 29,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/PROTEINS_full.zip',
		'domain': 'bioinformatics',
	},

	### computer vision
	'COIL-DEL': {
		'database': 'tudataset',
		'reference': '[16,18]',
		'dataset_size': 3900,
		'class_number': 100,
		'task_type': 'classification',
		'ave_node_num': 21.54,
		'ave_edge_num': 54.24,
		'node_labeled': False,
		'edge_labeled': True,
		'node_attr_dim': 2,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/COIL-DEL.zip',
		'domain': 'computer vision',
	},
	'COIL-RAG': {
		'database': 'tudataset',
		'reference': '[16,18]',
		'dataset_size': 3900,
		'class_number': 100,
		'task_type': 'classification',
		'ave_node_num': 3.01,
		'ave_edge_num': 3.02,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 64,
		'geometry': None,
		'edge_attr_dim': 1,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/COIL-RAG.zip',
		'domain': 'computer vision',
	},
	'Cuneiform': {
		'database': 'tudataset',
		'reference': '[25]',
		'dataset_size': 267,
		'class_number': 30,
		'task_type': 'classification',
		'ave_node_num': 21.27,
		'ave_edge_num': 44.8,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 3,
		'geometry': '3D',
		'edge_attr_dim': 2,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Cuneiform.zip',
		'domain': 'computer vision',
	},
	'Fingerprint': {
		'database': 'tudataset',
		'reference': '[16,19]',
		'dataset_size': 2800,
		'class_number': 4,
		'task_type': 'classification',
		'ave_node_num': 5.42,
		'ave_edge_num': 4.42,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 2,
		'geometry': '2D',
		'edge_attr_dim': 2,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Fingerprint.zip',
		'domain': 'computer vision',
	},
	'FIRSTMM_DB': {
		'database': 'tudataset',
		'reference': '[11,12,13]',
		'dataset_size': 41,
		'class_number': 11,
		'task_type': 'classification',
		'ave_node_num': 1377.27,
		'ave_edge_num': 3074.1,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 1,
		'geometry': None,
		'edge_attr_dim': 2,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/FIRSTMM_DB.zip',
		'domain': 'computer vision',
	},
	'Letter-high': {
		'database': 'tudataset',
		'reference': '[16]',
		'dataset_size': 2250,
		'class_number': 15,
		'task_type': 'classification',
		'ave_node_num': 4.67,
		'ave_edge_num': 4.5,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 2,
		'geometry': '2D',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Letter-high.zip',
		'domain': 'computer vision',
	},
	'Letter-low': {
		'database': 'tudataset',
		'reference': '[16]',
		'dataset_size': 2250,
		'class_number': 15,
		'task_type': 'classification',
		'ave_node_num': 4.68,
		'ave_edge_num': 3.13,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 2,
		'geometry': '2D',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Letter-low.zip',
		'domain': 'computer vision',
	},
	'Letter-med': {
		'database': 'tudataset',
		'reference': '[16]',
		'dataset_size': 2250,
		'class_number': 15,
		'task_type': 'classification',
		'ave_node_num': 4.67,
		'ave_edge_num': 4.5,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 2,
		'geometry': '2D',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Letter-med.zip',
		'domain': 'computer vision',
	},
	'MSRC_9': {
		'database': 'tudataset',
		'reference': '[13]',
		'dataset_size': 221,
		'class_number': 8,
		'task_type': 'classification',
		'ave_node_num': 40.58,
		'ave_edge_num': 97.94,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/MSRC_9.zip',
		'domain': 'computer vision',
	},
	'MSRC_21': {
		'database': 'tudataset',
		'reference': '[13]',
		'dataset_size': 563,
		'class_number': 20,
		'task_type': 'classification',
		'ave_node_num': 77.52,
		'ave_edge_num': 198.32,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/MSRC_21.zip',
		'domain': 'computer vision',
	},
	'MSRC_21C': {
		'database': 'tudataset',
		'reference': '[13]',
		'dataset_size': 209,
		'class_number': 20,
		'task_type': 'classification',
		'ave_node_num': 40.28,
		'ave_edge_num': 96.6,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/MSRC_21C.zip',
		'domain': 'computer vision',
	},

	### social networks
	'COLLAB': {
		'database': 'tudataset',
		'reference': '[14]',
		'dataset_size': 5000,
		'class_number': 3,
		'task_type': 'classification',
		'ave_node_num': 74.49,
		'ave_edge_num': 2457.78,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/COLLAB.zip',
		'domain': 'social networks',
	},
	'dblp_ct1': {
		'database': 'tudataset',
		'reference': '[32]',
		'dataset_size': 755,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 52.87,
		'ave_edge_num': 320.09,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 'temporal',
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/dblp_ct1.zip',
		'domain': 'social networks',
	},
	'dblp_ct2': {
		'database': 'tudataset',
		'reference': '[32]',
		'dataset_size': 755,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 52.87,
		'ave_edge_num': 320.09,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 'temporal',
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/dblp_ct2.zip',
		'domain': 'social networks',
	},
	'DBLP_v1': {
		'database': 'tudataset',
		'reference': '[26]',
		'dataset_size': 19456,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 10.48,
		'ave_edge_num': 19.65,
		'node_labeled': True,
		'edge_labeled': True,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/DBLP_v1.zip',
		'domain': 'social networks',
	},
	'deezer_ego_nets': {
		'database': 'tudataset',
		'reference': '[30]',
		'dataset_size': 9629,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 23.49,
		'ave_edge_num': 65.25,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/deezer_ego_nets.zip',
		'domain': 'social networks',
	},
	'facebook_ct1': {
		'database': 'tudataset',
		'reference': '[32]',
		'dataset_size': 995,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 95.72,
		'ave_edge_num': 269.01,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 'temporal',
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/facebook_ct1.zip',
		'domain': 'social networks',
	},
	'facebook_ct2': {
		'database': 'tudataset',
		'reference': '[32]',
		'dataset_size': 995,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 95.72,
		'ave_edge_num': 269.01,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 'temporal',
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/facebook_ct2.zip',
		'domain': 'social networks',
	},
	'github_stargazers': {
		'database': 'tudataset',
		'reference': '[30]',
		'dataset_size': 12725,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 113.79,
		'ave_edge_num': 234.64,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/github_stargazers.zip',
		'domain': 'social networks',
	},
	'highschool_ct1': {
		'database': 'tudataset',
		'reference': '[32]',
		'dataset_size': 180,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 52.32,
		'ave_edge_num': 544.81,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 'temporal',
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/highschool_ct1.zip',
		'domain': 'social networks',
	},
	'highschool_ct2': {
		'database': 'tudataset',
		'reference': '[32]',
		'dataset_size': 180,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 52.32,
		'ave_edge_num': 544.81,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 'temporal',
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/highschool_ct2.zip',
		'domain': 'social networks',
	},
	'IMDB-BINARY': {
		'database': 'tudataset',
		'reference': '[14]',
		'dataset_size': 1000,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 19.77,
		'ave_edge_num': 96.53,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': '',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/IMDB-BINARY.zip',
		'domain': 'social networks',
	},
	'IMDB-MULTI': {
		'database': 'tudataset',
		'reference': '[14]',
		'dataset_size': 1500,
		'class_number': 3,
		'task_type': 'classification',
		'ave_node_num': 13.0,
		'ave_edge_num': 65.94,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': '',
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/IMDB-MULTI.zip',
		'domain': 'social networks',
	},
	'infectious_ct1': {
		'database': 'tudataset',
		'reference': '[32]',
		'dataset_size': 200,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 50.0,
		'ave_edge_num': 459.72,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 'temporal',
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/infectious_ct1.zip',
		'domain': 'social networks',
	},
	'infectious_ct2': {
		'database': 'tudataset',
		'reference': '[32]',
		'dataset_size': 200,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 50.0,
		'ave_edge_num': 459.72,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 'temporal',
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/infectious_ct2.zip',
		'domain': 'social networks',
	},
	'mit_ct1': {
		'database': 'tudataset',
		'reference': '[32]',
		'dataset_size': 97,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 20.0,
		'ave_edge_num': 1469.15,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 'temporal',
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/mit_ct1.zip',
		'domain': 'social networks',
	},
	'mit_ct2': {
		'database': 'tudataset',
		'reference': '[32]',
		'dataset_size': 97,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 20.0,
		'ave_edge_num': 1469.15,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 'temporal',
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/mit_ct2.zip',
		'domain': 'social networks',
	},
	'REDDIT-BINARY': {
		'database': 'tudataset',
		'reference': '[14]',
		'dataset_size': 2000,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 429.63,
		'ave_edge_num': 497.75,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/REDDIT-BINARY.zip',
		'domain': 'social networks',
	},
	'REDDIT-MULTI-5K': {
		'database': 'tudataset',
		'reference': '[14]',
		'dataset_size': 4999,
		'class_number': 5,
		'task_type': 'classification',
		'ave_node_num': 508.52,
		'ave_edge_num': 594.87,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/REDDIT-MULTI-5K.zip',
		'domain': 'social networks',
	},
	'REDDIT-MULTI-12K': {
		'database': 'tudataset',
		'reference': '[14]',
		'dataset_size': 11929,
		'class_number': 11,
		'task_type': 'classification',
		'ave_node_num': 391.41,
		'ave_edge_num': 456.89,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/REDDIT-MULTI-12K.zip',
		'domain': 'social networks',
	},
	'reddit_threads': {
		'database': 'tudataset',
		'reference': '[30]',
		'dataset_size': 203088,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 23.93,
		'ave_edge_num': 24.99,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/reddit_threads.zip',
		'domain': 'social networks',
	},
	'tumblr_ct1': {
		'database': 'tudataset',
		'reference': '[32]',
		'dataset_size': 373,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 53.11,
		'ave_edge_num': 199.78,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 'temporal',
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/tumblr_ct1.zip',
		'domain': 'social networks',
	},
	'tumblr_ct2': {
		'database': 'tudataset',
		'reference': '[32]',
		'dataset_size': 373,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 53.11,
		'ave_edge_num': 199.78,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 'temporal',
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/tumblr_ct2.zip',
		'domain': 'social networks',
	},
	'twitch_egos': {
		'database': 'tudataset',
		'reference': '[30]',
		'dataset_size': 127094,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 29.67,
		'ave_edge_num': 86.59,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/twitch_egos.zip',
		'domain': 'social networks',
	},
	'TWITTER-Real-Graph-Partial': {
		'database': 'tudataset',
		'reference': '[26]',
		'dataset_size': 144033,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 4.03,
		'ave_edge_num': 4.98,
		'node_labeled': True,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 1,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/TWITTER-Real-Graph-Partial.zip',
		'domain': 'social networks',
	},

	### synthetic
	'COLORS-3': {
		'database': 'tudataset',
		'reference': '[27]',
		'dataset_size': 10500,
		'class_number': 11,
		'task_type': 'classification',
		'ave_node_num': 61.31,
		'ave_edge_num': 91.03,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 4,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/COLORS-3.zip',
		'domain': 'synthetic',
	},
	'SYNTHETIC': {
		'database': 'tudataset',
		'reference': '[3]',
		'dataset_size': 300,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 100.0,
		'ave_edge_num': 196.0,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 1,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/SYNTHETIC.zip',
		'domain': 'synthetic',
	},
	'SYNTHETICnew': {
		'database': 'tudataset',
		'reference': '[3,10]',
		'dataset_size': 300,
		'class_number': 2,
		'task_type': 'classification',
		'ave_node_num': 100.0,
		'ave_edge_num': 196.25,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 1,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/SYNTHETICnew.zip',
		'domain': 'synthetic',
	},
	'Synthie': {
		'database': 'tudataset',
		'reference': '[21]',
		'dataset_size': 400,
		'class_number': 4,
		'task_type': 'classification',
		'ave_node_num': 95.0,
		'ave_edge_num': 172.93,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 15,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/Synthie.zip',
		'domain': 'synthetic',
	},
	'TRIANGLES': {
		'database': 'tudataset',
		'reference': '[27]',
		'dataset_size': 45000,
		'class_number': 10,
		'task_type': 'classification',
		'ave_node_num': 20.85,
		'ave_edge_num': 32.74,
		'node_labeled': False,
		'edge_labeled': False,
		'node_attr_dim': 0,
		'geometry': None,
		'edge_attr_dim': 0,
		'url': 'https://www.chrsmrrs.com/graphkerneldatasets/TRIANGLES.zip',
		'domain': 'synthetic',
	},
}

DATASET_META = {**GREYC_META, **IAM_META, **TUDataset_META}


def list_of_databases():
	"""List names of all databases.

	Returns
	-------
	list
		The list of all databases.
	"""
	return [i for i in DATABASES]


def list_of_datasets():
	"""List names of all datasets.

	Returns
	-------
	list
		The list of all datasets.
	"""
	return [i for i in DATASET_META]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 22 11:31:26 2020

@author: ljia
"""
import numpy as np
from gklearn.utils import dummy_node, undefined_node


class NodeMap(object):
	
	def __init__(self, num_nodes_g, num_nodes_h):
		self._forward_map = [undefined_node()] * num_nodes_g
		self._backward_map = [undefined_node()] * num_nodes_h
		self._induced_cost = np.inf
		
		
	def clear(self):
		"""
	/*!
	 * @brief Clears the node map.
	 */
		"""
		self._forward_map = [undefined_node() for i in range(len(self._forward_map))]
		self._backward_map = [undefined_node() for i in range(len(self._backward_map))]
		
		
	def num_source_nodes(self):
		return len(self._forward_map)
	
	
	def num_target_nodes(self):
		return len(self._backward_map)
	
	
	def image(self, node):
		if node < len(self._forward_map):
			return self._forward_map[node]
		else:
			raise Exception('The node with ID ', str(node), ' is not contained in the source nodes of the node map.')
		return undefined_node()
	
	
	def pre_image(self, node):
		if node < len(self._backward_map):
			return self._backward_map[node]
		else:
			raise Exception('The node with ID ', str(node), ' is not contained in the target nodes of the node map.')
		return undefined_node()
	
	
	def as_relation(self, relation):
		relation.clear()
		for i in range(0, len(self._forward_map)):
			k = self._forward_map[i]
			if k != undefined_node():
				relation.append(tuple((i, k)))
		for k in range(0, len(self._backward_map)):
			i = self._backward_map[k]
			if i == dummy_node():
				relation.append(tuple((i, k)))
	
	
	def add_assignment(self, i, k):
		if i != dummy_node():
			if i < len(self._forward_map):
				self._forward_map[i] = k
			else:
				raise Exception('The node with ID ', str(i), ' is not contained in the source nodes of the node map.')
		if k != dummy_node():
			if k < len(self._backward_map):
				self._backward_map[k] = i
			else:
				raise Exception('The node with ID ', str(k), ' is not contained in the target nodes of the node map.')
	
	
	def set_induced_cost(self, induced_cost):
		self._induced_cost = induced_cost
		
		
	def induced_cost(self):
		return self._induced_cost
	
	
	@property
	def forward_map(self):
		return self._forward_map

	@forward_map.setter
	def forward_map(self, value):
		self._forward_map = value	
		
		
	@property
	def backward_map(self):
		return self._backward_map

	@backward_map.setter
	def backward_map(self, value):
		self._backward_map = value	


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Sep 11 18:10:06 2020

@author: ljia
"""
import numpy as np
import networkx as nx
import random


class GraphSynthesizer(object):
	
	
	def __init__(self, g_type=None, *args, **kwargs):
		if g_type == 'unified':
			self._graphs = self.unified_graphs(*args, *kwargs)
		else:
			self._graphs = None
	
	
	def random_graph(self, num_nodes, num_edges, num_node_labels=0, num_edge_labels=0, seed=None, directed=False, max_num_edges=None, all_edges=None):
		g = nx.Graph()
		if num_node_labels > 0:
			node_labels = np.random.randint(0, high=num_node_labels, size=num_nodes)
			for i in range(0, num_nodes):
				g.add_node(str(i), atom=node_labels[i]) # @todo: update "atom".
		else:
			for i in range(0, num_nodes):
				g.add_node(str(i))

		if num_edge_labels > 0:
			edge_labels = np.random.randint(0, high=num_edge_labels, size=num_edges)				
			for idx, i in enumerate(random.sample(range(0, max_num_edges), num_edges)):
				node1, node2 = all_edges[i]
				g.add_edge(str(node1), str(node2), bond_type=edge_labels[idx])  # @todo: update "bond_type".
		else:
			for i in random.sample(range(0, max_num_edges), num_edges):
				node1, node2 = all_edges[i]
				g.add_edge(str(node1), str(node2))
		
		return g
	
	
	def unified_graphs(self, num_graphs=1000, num_nodes=20, num_edges=40, num_node_labels=0, num_edge_labels=0, seed=None, directed=False):
		max_num_edges = int((num_nodes - 1) * num_nodes / 2)
		if num_edges > max_num_edges:
			raise Exception('Too many edges.')
		all_edges = [(i, j) for i in range(0, num_nodes) for j in range(i + 1, num_nodes)] # @todo: optimize. No directed graphs.
			
		graphs = []
		for idx in range(0, num_graphs):		
			graphs.append(self.random_graph(num_nodes, num_edges, num_node_labels=num_node_labels, num_edge_labels=num_edge_labels, seed=seed, directed=directed, max_num_edges=max_num_edges, all_edges=all_edges))
			
		return graphs
	
	
	@property
	def graphs(self):
		return self._graphs


""" Utilities function to manage graph files
"""
from os.path import dirname, splitext


class DataLoader():


	def __init__(self, filename, filename_targets=None, gformat=None, **kwargs):
		"""Read graph data from filename and load them as NetworkX graphs.

		Parameters
		----------
		filename : string
			The name of the file from where the dataset is read.
		filename_targets : string
			The name of file of the targets corresponding to graphs.

		Notes
		-----
		This function supports following graph dataset formats:

		'ds': load data from .ds file. See comments of function loadFromDS for a example.

		'cxl': load data from Graph eXchange Language file (.cxl file). See
		`here <http://www.gupro.de/GXL/Introduction/background.html>`__ for detail.

		'sdf': load data from structured data file (.sdf file). See
		`here <http://www.nonlinear.com/progenesis/sdf-studio/v0.9/faq/sdf-file-format-guidance.aspx>`__
		for details.

		'mat': Load graph data from a MATLAB (up to version 7.1) .mat file. See
		README in `downloadable file <http://mlcb.is.tuebingen.mpg.de/Mitarbeiter/Nino/WL/>`__
		for details.

		'txt': Load graph data from the TUDataset. See
		`here <https://ls11-www.cs.tu-dortmund.de/staff/morris/graphkerneldatasets>`__
		for details. Note here filename is the name of either .txt file in
		the dataset directory.
		"""
		if isinstance(filename, str):
			extension = splitext(filename)[1][1:]
		else: # filename is a list of files.
			extension = splitext(filename[0])[1][1:]

		if extension == "ds":
			self._graphs, self._targets, self._label_names = self.load_from_ds(filename, filename_targets)
		elif extension == "cxl":
			dir_dataset = kwargs.get('dirname_dataset', None)
			self._graphs, self._targets, self._label_names = self.load_from_xml(filename, dir_dataset)
		elif extension == 'xml':
			dir_dataset = kwargs.get('dirname_dataset', None)
			self._graphs, self._targets, self._label_names = self.load_from_xml(filename, dir_dataset)
		elif extension == "mat":
			order = kwargs.get('order')
			self._graphs, self._targets, self._label_names = self.load_mat(filename, order)
		elif extension == 'txt':
			if gformat is None:
				self._graphs, self._targets, self._label_names = self.load_tud(filename)
			elif gformat == 'cml':
				self._graphs, self._targets, self._label_names = self.load_from_ds(filename, filename_targets)

		else:
			raise ValueError('The input file with the extension ".', extension, '" is not supported. The supported extensions includes: ".ds", ".cxl", ".xml", ".mat", ".txt".')


	def load_from_ds(self, filename, filename_targets):
		"""Load data from .ds file.

		Possible graph formats include:

		'.ct': see function load_ct for detail.

		'.gxl': see dunction load_gxl for detail.

		Note these graph formats are checked automatically by the extensions of
		graph files.
		"""
		if isinstance(filename, str):
			dirname_dataset = dirname(filename)
			with open(filename) as f:
				content = f.read().splitlines()
		else: # filename is a list of files.
			dirname_dataset = dirname(filename[0])
			content = []
			for fn in filename:
				with open(fn) as f:
					content += f.read().splitlines()
			# to remove duplicate file names.

		data = []
		y = []
		label_names = {'node_labels': [], 'edge_labels': [], 'node_attrs': [], 'edge_attrs': []}
		content = [line for line in content if not line.endswith('.ds')] # Alkane
		content = [line for line in content if not line.startswith('#')] # Acyclic
		extension = splitext(content[0].split(' ')[0])[1][1:]
		if extension == 'ct':
			load_file_fun = self.load_ct
		elif extension == 'gxl' or extension == 'sdf': # @todo: .sdf not tested yet.
			load_file_fun = self.load_gxl
		elif extension == 'cml': # dataset "Chiral"
			load_file_fun = self.load_cml

		if filename_targets is None or filename_targets == '':
			for i in range(0, len(content)):
				tmp = content[i].split(' ')
				# remove the '#'s in file names
				g, l_names = load_file_fun(dirname_dataset + '/' + tmp[0].replace('#', '', 1))
				data.append(g)
				self._append_label_names(label_names, l_names) # @todo: this is so redundant.
				y.append(float(tmp[1]))
		else:  # targets in a seperate file
			for i in range(0, len(content)):
				tmp = content[i]
				# remove the '#'s in file names
				g, l_names = load_file_fun(dirname_dataset + '/' + tmp.replace('#', '', 1))
				data.append(g)
				self._append_label_names(label_names, l_names)

			with open(filename_targets) as fnt:
				content_y = fnt.read().splitlines()
			# assume entries in filename and filename_targets have the same order.
			for item in content_y:
				tmp = item.split(' ')
				# assume the 3rd entry in a line is y (for Alkane dataset)
				y.append(float(tmp[2]))

		return data, y, label_names


	def load_from_xml(self, filename, dir_dataset=None):
		import xml.etree.ElementTree as ET

		def load_one_file(filename, data, y, label_names):
			tree = ET.parse(filename)
			root = tree.getroot()
			for graph in root.iter('graph') if root.find('graph') is not None else root.iter('print'): # "graph" for ... I forgot; "print" for datasets GREC and Web.
				mol_filename = graph.attrib['file']
				mol_class = graph.attrib['class']
				g, l_names = self.load_gxl(dir_dataset + '/' + mol_filename)
				data.append(g)
				self._append_label_names(label_names, l_names)
				y.append(mol_class)

		data = []
		y = []
		label_names = {'node_labels': [], 'edge_labels': [], 'node_attrs': [], 'edge_attrs': []}

		if isinstance(filename, str):
			if dir_dataset is not None:
				dir_dataset = dir_dataset
			else:
				dir_dataset = dirname(filename)
			load_one_file(filename, data, y, label_names)


		else: # filename is a list of files.
			if dir_dataset is not None:
				dir_dataset = dir_dataset
			else:
				dir_dataset = dirname(filename[0])

			for fn in filename:
				load_one_file(fn, data, y, label_names)

		return data, y, label_names


	def load_mat(self, filename, order): # @todo: need to be updated (auto order) or deprecated.
		"""Load graph data from a MATLAB (up to version 7.1) .mat file.

		Notes
		------
		A MAT file contains a struct array containing graphs, and a column vector lx containing a class label for each graph.
		Check README in `downloadable file <http://mlcb.is.tuebingen.mpg.de/Mitarbeiter/Nino/WL/>`__ for detailed structure.
		"""
		from scipy.io import loadmat
		import numpy as np
		import networkx as nx
		data = []
		content = loadmat(filename)
		for key, value in content.items():
			if key[0] == 'l':  # class label
				y = np.transpose(value)[0].tolist()
			elif key[0] != '_':
				# if adjacency matrix is not compressed / edge label exists
				if order[1] == 0:
					for i, item in enumerate(value[0]):
						g = nx.Graph(name=i)  # set name of the graph
						nl = np.transpose(item[order[3]][0][0][0])  # node label
						for index, label in enumerate(nl[0]):
							g.add_node(index, label_1=str(label))
						el = item[order[4]][0][0][0]  # edge label
						for edge in el:
							g.add_edge(edge[0] - 1, edge[1] - 1, label_1=str(edge[2]))
						data.append(g)
				else:
					for i, item in enumerate(value[0]):
						g = nx.Graph(name=i)  # set name of the graph
						nl = np.transpose(item[order[3]][0][0][0])  # node label
						for index, label in enumerate(nl[0]):
							g.add_node(index, label_1=str(label))
						sam = item[order[0]]  # sparse adjacency matrix
						index_no0 = sam.nonzero()
						for col, row in zip(index_no0[0], index_no0[1]):
							g.add_edge(col, row)
						data.append(g)

		label_names = {'node_labels': ['label_1'], 'edge_labels': [], 'node_attrs': [], 'edge_attrs': []}
		if order[1] == 0:
			label_names['edge_labels'].append('label_1')

		return data, y, label_names


	def load_tud(self, filename):
		"""Load graph data from TUD dataset files.

		Notes
		------
		The graph data is loaded from separate files.
		Check README in `downloadable file <http://tiny.cc/PK_MLJ_data>`__, 2018 for detailed structure.
		"""
		import networkx as nx
		from os import listdir
		from os.path import dirname, basename


		def get_infos_from_readme(frm): # @todo: add README (cuniform), maybe node/edge label maps.
			"""Get information from DS_label_readme.txt file.
			"""

			def get_label_names_from_line(line):
				"""Get names of labels/attributes from a line.
				"""
				str_names = line.split('[')[1].split(']')[0]
				names = str_names.split(',')
				names = [attr.strip() for attr in names]
				return names


			def get_class_label_map(label_map_strings):
				label_map = {}
				for string in label_map_strings:
					integer, label = string.split('\t')
					label_map[int(integer.strip())] = label.strip()
				return label_map


			label_names = {'node_labels': [], 'node_attrs': [],
						   'edge_labels': [], 'edge_attrs': []}
			class_label_map = None
			class_label_map_strings = []
			with open(frm) as rm:
				content_rm = rm.read().splitlines()
			i = 0
			while i < len(content_rm):
				line = content_rm[i].strip()
				# get node/edge labels and attributes.
				if line.startswith('Node labels:'):
					label_names['node_labels'] = get_label_names_from_line(line)
				elif line.startswith('Node attributes:'):
					label_names['node_attrs'] = get_label_names_from_line(line)
				elif line.startswith('Edge labels:'):
					label_names['edge_labels'] = get_label_names_from_line(line)
				elif line.startswith('Edge attributes:'):
					label_names['edge_attrs'] = get_label_names_from_line(line)
				# get class label map.
				elif line.startswith('Class labels were converted to integer values using this map:'):
					i += 2
					line = content_rm[i].strip()
					while line != '' and i < len(content_rm):
						class_label_map_strings.append(line)
						i += 1
						line = content_rm[i].strip()
					class_label_map = get_class_label_map(class_label_map_strings)
				i += 1

			return label_names, class_label_map


		# get dataset name.
		dirname_dataset = dirname(filename)
		filename = basename(filename)
		fn_split = filename.split('_A')
		ds_name = fn_split[0].strip()

		# load data file names
		for name in listdir(dirname_dataset):
			if ds_name + '_A' in name:
				fam = dirname_dataset + '/' + name
			elif ds_name + '_graph_indicator' in name:
				fgi = dirname_dataset + '/' + name
			elif ds_name + '_graph_labels' in name:
				fgl = dirname_dataset + '/' + name
			elif ds_name + '_node_labels' in name:
				fnl = dirname_dataset + '/' + name
			elif ds_name + '_edge_labels' in name:
				fel = dirname_dataset + '/' + name
			elif ds_name + '_edge_attributes' in name:
				fea = dirname_dataset + '/' + name
			elif ds_name + '_node_attributes' in name:
				fna = dirname_dataset + '/' + name
			elif ds_name + '_graph_attributes' in name:
				fga = dirname_dataset + '/' + name
			elif ds_name + '_label_readme' in name:
				frm = dirname_dataset + '/' + name
			# this is supposed to be the node attrs, make sure to put this as the last 'elif'
			elif ds_name + '_attributes' in name:
				fna = dirname_dataset + '/' + name

		# get labels and attributes names.
		if 'frm' in locals():
			label_names, class_label_map = get_infos_from_readme(frm)
		else:
			label_names = {'node_labels': [], 'node_attrs': [],
						   'edge_labels': [], 'edge_attrs': []}
			class_label_map = None

		with open(fgi) as gi:
			content_gi = gi.read().splitlines()  # graph indicator
		with open(fam) as am:
			content_am = am.read().splitlines()  # adjacency matrix

		# load targets.
		if 'fgl' in locals():
			with open(fgl) as gl:
				content_targets = gl.read().splitlines()  # targets (classification)
			targets = [float(i) for i in content_targets]
		elif 'fga' in locals():
			with open(fga) as ga:
				content_targets = ga.read().splitlines()  # targets (regression)
			targets = [int(i) for i in content_targets]
		else:
			exp_msg = 'Can not find targets file. Please make sure there is a "', ds_name, '_graph_labels.txt" or "', ds_name, '_graph_attributes.txt"', 'file in your dataset folder.'
			raise Exception(exp_msg)
		if class_label_map is not None:
			targets = [class_label_map[t] for t in targets]

		# create graphs and add nodes
		data = [nx.Graph(name=str(i)) for i in range(0, len(content_targets))]
		if 'fnl' in locals():
			with open(fnl) as nl:
				content_nl = nl.read().splitlines()  # node labels
			for idx, line in enumerate(content_gi):
				# transfer to int first in case of unexpected blanks
				data[int(line) - 1].add_node(idx)
				labels = [l.strip() for l in content_nl[idx].split(',')]
				if label_names['node_labels'] == []: # @todo: need fix bug.
					for i, label in enumerate(labels):
						l_name = 'label_' + str(i)
						data[int(line) - 1].nodes[idx][l_name] = label
						label_names['node_labels'].append(l_name)
				else:
					for i, l_name in enumerate(label_names['node_labels']):
						data[int(line) - 1].nodes[idx][l_name] = labels[i]
		else:
			for i, line in enumerate(content_gi):
				data[int(line) - 1].add_node(i)

		# add edges
		for line in content_am:
			tmp = line.split(',')
			n1 = int(tmp[0]) - 1
			n2 = int(tmp[1]) - 1
			# ignore edge weight here.
			g = int(content_gi[n1]) - 1
			data[g].add_edge(n1, n2)

		# add edge labels
		if 'fel' in locals():
			with open(fel) as el:
				content_el = el.read().splitlines()
			for idx, line in enumerate(content_el):
				labels = [l.strip() for l in line.split(',')]
				n = [int(i) - 1 for i in content_am[idx].split(',')]
				g = int(content_gi[n[0]]) - 1
				if label_names['edge_labels'] == []:
					for i, label in enumerate(labels):
						l_name = 'label_' + str(i)
						data[g].edges[n[0], n[1]][l_name] = label
						label_names['edge_labels'].append(l_name)
				else:
					for i, l_name in enumerate(label_names['edge_labels']):
						data[g].edges[n[0], n[1]][l_name] = labels[i]

		# add node attributes
		if 'fna' in locals():
			with open(fna) as na:
				content_na = na.read().splitlines()
			for idx, line in enumerate(content_na):
				attrs = [a.strip() for a in line.split(',')]
				g = int(content_gi[idx]) - 1
				if label_names['node_attrs'] == []:
					for i, attr in enumerate(attrs):
						a_name = 'attr_' + str(i)
						data[g].nodes[idx][a_name] = attr
						label_names['node_attrs'].append(a_name)
				else:
					for i, a_name in enumerate(label_names['node_attrs']):
						data[g].nodes[idx][a_name] = attrs[i]

		# add edge attributes
		if 'fea' in locals():
			with open(fea) as ea:
				content_ea = ea.read().splitlines()
			for idx, line in enumerate(content_ea):
				attrs = [a.strip() for a in line.split(',')]
				n = [int(i) - 1 for i in content_am[idx].split(',')]
				g = int(content_gi[n[0]]) - 1
				if label_names['edge_attrs'] == []:
					for i, attr in enumerate(attrs):
						a_name = 'attr_' + str(i)
						data[g].edges[n[0], n[1]][a_name] = attr
						label_names['edge_attrs'].append(a_name)
				else:
					for i, a_name in enumerate(label_names['edge_attrs']):
						data[g].edges[n[0], n[1]][a_name] = attrs[i]

		return data, targets, label_names


	def load_ct(self, filename): # @todo: this function is only tested on CTFile V2000; header not considered; only simple cases (atoms and bonds are considered.)
		"""load data from a Chemical Table (.ct) file.

		Notes
		------
		a typical example of data in .ct is like this:

		3 2  <- number of nodes and edges

		0.0000	0.0000	0.0000 C <- each line describes a node (x,y,z + label)

		0.0000	0.0000	0.0000 C

		0.0000	0.0000	0.0000 O

		1  3  1  1 <- each line describes an edge : to, from, bond type, bond stereo

		2  3  1  1

		Check `CTFile Formats file <https://www.google.com/url?sa=t&rct=j&q=&esrc=s&source=web&cd=10&ved=2ahUKEwivhaSdjsTlAhVhx4UKHczHA8gQFjAJegQIARAC&url=https%3A%2F%2Fwww.daylight.com%2Fmeetings%2Fmug05%2FKappler%2Fctfile.pdf&usg=AOvVaw1cDNrrmMClkFPqodlF2inS>`__
		for detailed format discription.
		"""
		import networkx as nx
		from os.path import basename
		g = nx.Graph()
		with open(filename) as f:
			content = f.read().splitlines()
			g = nx.Graph(name=str(content[0]), filename=basename(filename))  # set name of the graph

			# read the counts line.
			tmp = content[1].split(' ')
			tmp = [x for x in tmp if x != '']
			nb_atoms = int(tmp[0].strip())  # number of atoms
			nb_bonds = int(tmp[1].strip())  # number of bonds
			count_line_tags = ['number_of_atoms', 'number_of_bonds', 'number_of_atom_lists', '', 'chiral_flag', 'number_of_stext_entries', '', '', '', '', 'number_of_properties', 'CT_version']
			i = 0
			while i < len(tmp):
				if count_line_tags[i] != '': # if not obsoleted
					g.graph[count_line_tags[i]] = tmp[i].strip()
				i += 1

			# read the atom block.
			atom_tags = ['x', 'y', 'z', 'atom_symbol', 'mass_difference', 'charge', 'atom_stereo_parity', 'hydrogen_count_plus_1', 'stereo_care_box', 'valence', 'h0_designator', '', '', 'atom_atom_mapping_number', 'inversion_retention_flag', 'exact_change_flag']
			for i in range(0, nb_atoms):
				tmp = content[i + 2].split(' ')
				tmp = [x for x in tmp if x != '']
				g.add_node(i)
				j = 0
				while j < len(tmp):
					if atom_tags[j] != '':
						g.nodes[i][atom_tags[j]] = tmp[j].strip()
					j += 1

			# read the bond block.
			bond_tags = ['first_atom_number', 'second_atom_number', 'bond_type', 'bond_stereo', '', 'bond_topology', 'reacting_center_status']
			for i in range(0, nb_bonds):
				tmp = content[i + g.number_of_nodes() + 2].split(' ')
				tmp = [x for x in tmp if x != '']
				n1, n2 = int(tmp[0].strip()) - 1, int(tmp[1].strip()) - 1
				g.add_edge(n1, n2)
				j = 2
				while j < len(tmp):
					if bond_tags[j] != '':
						g.edges[(n1, n2)][bond_tags[j]] = tmp[j].strip()
					j += 1

		# get label names.
		label_names = {'node_labels': [], 'edge_labels': [], 'node_attrs': [], 'edge_attrs': []}
		atom_symbolic = [0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, None, None, 1, 1, 1]
		for nd in g.nodes():
			for key in g.nodes[nd]:
				if atom_symbolic[atom_tags.index(key)] == 1:
					label_names['node_labels'].append(key)
				else:
					label_names['node_attrs'].append(key)
			break
		bond_symbolic = [None, None, 1, 1, None, 1, 1]
		for ed in g.edges():
			for key in g.edges[ed]:
				if bond_symbolic[bond_tags.index(key)] == 1:
					label_names['edge_labels'].append(key)
				else:
					label_names['edge_attrs'].append(key)
			break

		return g, label_names


	def load_gxl(self, filename): # @todo: directed graphs.
		from os.path import basename
		import networkx as nx
		import xml.etree.ElementTree as ET

		tree = ET.parse(filename)
		root = tree.getroot()
		index = 0
		g = nx.Graph(filename=basename(filename), name=root[0].attrib['id'])
		dic = {}  # used to retrieve incident nodes of edges
		for node in root.iter('node'):
			dic[node.attrib['id']] = index
			labels = {}
			for attr in node.iter('attr'): # for datasets "GREC" and "Monoterpens".
				labels[attr.attrib['name']] = attr[0].text
			for attr in node.iter('attribute'): # for dataset "Web".
				labels[attr.attrib['name']] = attr.attrib['value']
			g.add_node(index, **labels)
			index += 1

		for edge in root.iter('edge'):
			labels = {}
			for attr in edge.iter('attr'): # for datasets "GREC" and "Monoterpens".
				labels[attr.attrib['name']] = attr[0].text
			for attr in edge.iter('attribute'): # for dataset "Web".
				labels[attr.attrib['name']] = attr.attrib['value']
			g.add_edge(dic[edge.attrib['from']], dic[edge.attrib['to']], **labels)

		# get label names.
		label_names = {'node_labels': [], 'edge_labels': [], 'node_attrs': [], 'edge_attrs': []}
		# @todo: possible loss of label names if some nodes miss some labels.
		for node in root.iter('node'):
			for attr in node.iter('attr'): # for datasets "GREC" and "Monoterpens".
				if attr[0].tag == 'int' or attr.attrib['name'] == 'type': # @todo: this maybe wrong, and slow. "type" is for dataset GREC; "int" is for dataset "Monoterpens".
					label_names['node_labels'].append(attr.attrib['name'])
				else:
					label_names['node_attrs'].append(attr.attrib['name'])

			for attr in node.iter('attribute'): # for dataset "Web".
				label_names['node_attrs'].append(attr.attrib['name'])
				# @todo: is id useful in dataset "Web"? is "FREQUENCY" symbolic or not?
			break

		for edge in root.iter('edge'):
			for attr in edge.iter('attr'): # for datasets "GREC" and "Monoterpens".
				if attr[0].tag == 'int' or attr.attrib['name'] == 'frequency' or 'type' in attr.attrib['name']: # @todo: this maybe wrong, and slow. "frequency" and "type" are for dataset GREC; "int" is for dataset "Monoterpens".
					label_names['edge_labels'].append(attr.attrib['name'])
				else:
					label_names['edge_attrs'].append(attr.attrib['name'])

			for attr in edge.iter('attribute'): # for dataset "Web".
				label_names['edge_attrs'].append(attr.attrib['name'])
			break

		return g, label_names


	def load_cml(self, filename): # @todo: directed graphs.
		# @todo: what is "atomParity" and "bondStereo" in the data file?
		from os.path import basename
		import networkx as nx
		import xml.etree.ElementTree as ET

		xmlns = '{http://www.xml-cml.org/schema}' # @todo: why this has to be added?
		tree = ET.parse(filename)
		root = tree.getroot()
		index = 0
		if root.tag == xmlns + 'molecule':
			g_id = root.attrib['id']
		else:
			g_id = root.find(xmlns + 'molecule').attrib['id']
		g = nx.Graph(filename=basename(filename), name=g_id)
		dic = {}  # used to retrieve incident nodes of edges
		for atom in root.iter(xmlns + 'atom'):
			dic[atom.attrib['id']] = index
			labels = {}
			for key, val in atom.attrib.items():
				if key != 'id':
					labels[key] = val
			g.add_node(index, **labels)
			index += 1

		for bond in root.iter(xmlns + 'bond'):
			labels = {}
			for key, val in bond.attrib.items():
				if key != 'atomRefs2' and key != 'id': # "id" is in dataset "ACE".
					labels[key] = val
			n1, n2 = bond.attrib['atomRefs2'].strip().split(' ')
			g.add_edge(dic[n1], dic[n2], **labels)

		# get label names.
		label_names = {'node_labels': [], 'edge_labels': [], 'node_attrs': [], 'edge_attrs': []}
		# @todo: possible loss of label names if some nodes miss some labels.
		for key, val in g.nodes[0].items():
			try:
				float(val)
			except:
				label_names['node_labels'].append(key)
			else:
				if val.isdigit():
					label_names['node_labels'].append(key)
				else:
					label_names['node_attrs'].append(key)
		for _, _, attrs in g.edges(data=True):
			for key, val in attrs.items():
				try:
					float(val)
				except:
					label_names['edge_labels'].append(key)
				else:
					if val.isdigit():
						label_names['edge_labels'].append(key)
					else:
						label_names['edge_attrs'].append(key)
			break

		return g, label_names


	def _append_label_names(self, label_names, new_names):
		for key, val in label_names.items():
			label_names[key] += [name for name in new_names[key] if name not in val]


	@property
	def data(self):
		return self._graphs, self._targets, self._label_names


	@property
	def graphs(self):
		return self._graphs


	@property
	def targets(self):
		return self._targets


	@property
	def label_names(self):
		return self._label_names


class DataSaver():


	def __init__(self, graphs, targets=None, filename='gfile', gformat='gxl', group=None, **kwargs):
		"""Save list of graphs.
		"""
		import os
		dirname_ds = os.path.dirname(filename)
		if dirname_ds != '':
			dirname_ds += '/'
			os.makedirs(dirname_ds, exist_ok=True)

		if 'graph_dir' in kwargs:
			graph_dir = kwargs['graph_dir'] + '/'
			os.makedirs(graph_dir, exist_ok=True)
			del kwargs['graph_dir']
		else:
			graph_dir = dirname_ds

		if group == 'xml' and gformat == 'gxl':
			with open(filename + '.xml', 'w') as fgroup:
				fgroup.write("<?xml version=\"1.0\"?>")
				fgroup.write("\n<!DOCTYPE GraphCollection SYSTEM \"http://www.inf.unibz.it/~blumenthal/dtd/GraphCollection.dtd\">")
				fgroup.write("\n<GraphCollection>")
				for idx, g in enumerate(graphs):
					fname_tmp = "graph" + str(idx) + ".gxl"
					self.save_gxl(g, graph_dir + fname_tmp, **kwargs)
					fgroup.write("\n\t<graph file=\"" + fname_tmp + "\" class=\"" + str(targets[idx]) + "\"/>")
				fgroup.write("\n</GraphCollection>")
				fgroup.close()


	def save_gxl(self, graph, filename, method='default', node_labels=[], edge_labels=[], node_attrs=[], edge_attrs=[]):
		if method == 'default':
			gxl_file = open(filename, 'w')
			gxl_file.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n")
			gxl_file.write("<!DOCTYPE gxl SYSTEM \"http://www.gupro.de/GXL/gxl-1.0.dtd\">\n")
			gxl_file.write("<gxl xmlns:xlink=\"http://www.w3.org/1999/xlink\">\n")
			if 'name' in graph.graph:
				name = str(graph.graph['name'])
			else:
				name = 'dummy'
			gxl_file.write("<graph id=\"" + name + "\" edgeids=\"false\" edgemode=\"undirected\">\n")
			for v, attrs in graph.nodes(data=True):
				gxl_file.write("<node id=\"_" + str(v) + "\">")
				for l_name in node_labels:
					gxl_file.write("<attr name=\"" + l_name + "\"><int>" +
								   str(attrs[l_name]) + "</int></attr>")
				for a_name in node_attrs:
					gxl_file.write("<attr name=\"" + a_name + "\"><float>" +
								   str(attrs[a_name]) + "</float></attr>")
				gxl_file.write("</node>\n")
			for v1, v2, attrs in graph.edges(data=True):
				gxl_file.write("<edge from=\"_" + str(v1) + "\" to=\"_" + str(v2) + "\">")
				for l_name in edge_labels:
					gxl_file.write("<attr name=\"" + l_name + "\"><int>" +
								   str(attrs[l_name]) + "</int></attr>")
				for a_name in edge_attrs:
					gxl_file.write("<attr name=\"" + a_name + "\"><float>" +
								   str(attrs[a_name]) + "</float></attr>")
				gxl_file.write("</edge>\n")
			gxl_file.write("</graph>\n")
			gxl_file.write("</gxl>")
			gxl_file.close()
		elif method == 'benoit':
			import xml.etree.ElementTree as ET
			root_node = ET.Element('gxl')
			attr = dict()
			attr['id'] = str(graph.graph['name'])
			attr['edgeids'] = 'true'
			attr['edgemode'] = 'undirected'
			graph_node = ET.SubElement(root_node, 'graph', attrib=attr)

			for v in graph:
				current_node = ET.SubElement(graph_node, 'node', attrib={'id': str(v)})
				for attr in graph.nodes[v].keys():
					cur_attr = ET.SubElement(
						current_node, 'attr', attrib={'name': attr})
					cur_value = ET.SubElement(cur_attr,
											  graph.nodes[v][attr].__class__.__name__)
					cur_value.text = graph.nodes[v][attr]

			for v1 in graph:
				for v2 in graph[v1]:
					if (v1 < v2):  # Non oriented graphs
						cur_edge = ET.SubElement(
							graph_node,
							'edge',
							attrib={
								'from': str(v1),
								'to': str(v2)
							})
						for attr in graph[v1][v2].keys():
							cur_attr = ET.SubElement(
								cur_edge, 'attr', attrib={'name': attr})
							cur_value = ET.SubElement(
								cur_attr, graph[v1][v2][attr].__class__.__name__)
							cur_value.text = str(graph[v1][v2][attr])

			tree = ET.ElementTree(root_node)
			tree.write(filename)
		elif method == 'gedlib':
			# reference: https://github.com/dbblumenthal/gedlib/blob/master/data/generate_molecules.py#L22
	#		pass
			gxl_file = open(filename, 'w')
			gxl_file.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n")
			gxl_file.write("<!DOCTYPE gxl SYSTEM \"http://www.gupro.de/GXL/gxl-1.0.dtd\">\n")
			gxl_file.write("<gxl xmlns:xlink=\"http://www.w3.org/1999/xlink\">\n")
			gxl_file.write("<graph id=\"" + str(graph.graph['name']) + "\" edgeids=\"true\" edgemode=\"undirected\">\n")
			for v, attrs in graph.nodes(data=True):
				gxl_file.write("<node id=\"_" + str(v) + "\">")
				gxl_file.write("<attr name=\"" + "chem" + "\"><int>" + str(attrs['chem']) + "</int></attr>")
				gxl_file.write("</node>\n")
			for v1, v2, attrs in graph.edges(data=True):
				gxl_file.write("<edge from=\"_" + str(v1) + "\" to=\"_" + str(v2) + "\">")
				gxl_file.write("<attr name=\"valence\"><int>" + str(attrs['valence']) + "</int></attr>")
	#			gxl_file.write("<attr name=\"valence\"><int>" + "1" + "</int></attr>")
				gxl_file.write("</edge>\n")
			gxl_file.write("</graph>\n")
			gxl_file.write("</gxl>")
			gxl_file.close()
		elif method == 'gedlib-letter':
			# reference: https://github.com/dbblumenthal/gedlib/blob/master/data/generate_molecules.py#L22
			# and https://github.com/dbblumenthal/gedlib/blob/master/data/datasets/Letter/HIGH/AP1_0000.gxl
			gxl_file = open(filename, 'w')
			gxl_file.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n")
			gxl_file.write("<!DOCTYPE gxl SYSTEM \"http://www.gupro.de/GXL/gxl-1.0.dtd\">\n")
			gxl_file.write("<gxl xmlns:xlink=\"http://www.w3.org/1999/xlink\">\n")
			gxl_file.write("<graph id=\"" + str(graph.graph['name']) + "\" edgeids=\"false\" edgemode=\"undirected\">\n")
			for v, attrs in graph.nodes(data=True):
				gxl_file.write("<node id=\"_" + str(v) + "\">")
				gxl_file.write("<attr name=\"x\"><float>" + str(attrs['attributes'][0]) + "</float></attr>")
				gxl_file.write("<attr name=\"y\"><float>" + str(attrs['attributes'][1]) + "</float></attr>")
				gxl_file.write("</node>\n")
			for v1, v2, attrs in graph.edges(data=True):
				gxl_file.write("<edge from=\"_" + str(v1) + "\" to=\"_" + str(v2) + "\"/>\n")
			gxl_file.write("</graph>\n")
			gxl_file.write("</gxl>")
			gxl_file.close()


# def loadSDF(filename):
# 	"""load data from structured data file (.sdf file).

# 	Notes
# 	------
# 	A SDF file contains a group of molecules, represented in the similar way as in MOL format.
# 	Check `here <http://www.nonlinear.com/progenesis/sdf-studio/v0.9/faq/sdf-file-format-guidance.aspx>`__ for detailed structure.
# 	"""
# 	import networkx as nx
# 	from os.path import basename
# 	from tqdm import tqdm
# 	import sys
# 	data = []
# 	with open(filename) as f:
# 		content = f.read().splitlines()
# 		index = 0
# 		pbar = tqdm(total=len(content) + 1, desc='load SDF', file=sys.stdout)
# 		while index < len(content):
# 			index_old = index

# 			g = nx.Graph(name=content[index].strip())  # set name of the graph

# 			tmp = content[index + 3]
# 			nb_nodes = int(tmp[:3])  # number of the nodes
# 			nb_edges = int(tmp[3:6])  # number of the edges

# 			for i in range(0, nb_nodes):
# 				tmp = content[i + index + 4]
# 				g.add_node(i, atom=tmp[31:34].strip())

# 			for i in range(0, nb_edges):
# 				tmp = content[i + index + g.number_of_nodes() + 4]
# 				tmp = [tmp[i:i + 3] for i in range(0, len(tmp), 3)]
# 				g.add_edge(
# 					int(tmp[0]) - 1, int(tmp[1]) - 1, bond_type=tmp[2].strip())

# 			data.append(g)

# 			index += 4 + g.number_of_nodes() + g.number_of_edges()
# 			while content[index].strip() != '$$$$':  # seperator
# 				index += 1
# 			index += 1

# 			pbar.update(index - index_old)
# 		pbar.update(1)
# 		pbar.close()

# 	return data


# def load_from_cxl(filename):
# 	import xml.etree.ElementTree as ET
#
# 	dirname_dataset = dirname(filename)
# 	tree = ET.parse(filename)
# 	root = tree.getroot()
# 	data = []
# 	y = []
# 	for graph in root.iter('graph'):
# 		mol_filename = graph.attrib['file']
# 		mol_class = graph.attrib['class']
# 		data.append(load_gxl(dirname_dataset + '/' + mol_filename))
# 		y.append(mol_class)


if __name__ == '__main__':
#	### Load dataset from .ds file.
#	# .ct files.
#	ds = {'name': 'Alkane', 'dataset': '../../datasets/Alkane/dataset.ds',
#		'dataset_y': '../../datasets/Alkane/dataset_boiling_point_names.txt'}
#	Gn, y = loadDataset(ds['dataset'], filename_y=ds['dataset_y'])
# 	ds_file = '../../datasets/Acyclic/dataset_bps.ds'  # node symb
# 	Gn, targets, label_names = load_dataset(ds_file)
# 	ds_file = '../../datasets/MAO/dataset.ds' # node/edge symb
# 	Gn, targets, label_names = load_dataset(ds_file)
##	ds = {'name': 'PAH', 'dataset': '../../datasets/PAH/dataset.ds'} # unlabeled
##	Gn, y = loadDataset(ds['dataset'])
# 	print(Gn[1].graph)
# 	print(Gn[1].nodes(data=True))
# 	print(Gn[1].edges(data=True))
# 	print(targets[1])

# 	# .gxl file.
# 	ds_file = '../../datasets/monoterpenoides/dataset_10+.ds'  # node/edge symb
# 	Gn, y, label_names = load_dataset(ds_file)
# 	print(Gn[1].graph)
# 	print(Gn[1].nodes(data=True))
# 	print(Gn[1].edges(data=True))
# 	print(y[1])

	# .mat file.
	ds_file = '../../datasets/MUTAG_mat/MUTAG.mat'
	order = [0, 0, 3, 1, 2]
	gloader = DataLoader(ds_file, order=order)
	Gn, targets, label_names = gloader.data
	print(Gn[1].graph)
	print(Gn[1].nodes(data=True))
	print(Gn[1].edges(data=True))
	print(targets[1])

#	### Convert graph from one format to another.
#	# .gxl file.
#	import networkx as nx
#	ds = {'name': 'monoterpenoides',
#		  'dataset': '../../datasets/monoterpenoides/dataset_10+.ds'}  # node/edge symb
#	Gn, y = loadDataset(ds['dataset'])
#	y = [int(i) for i in y]
#	print(Gn[1].nodes(data=True))
#	print(Gn[1].edges(data=True))
#	print(y[1])
#	# Convert a graph to the proper NetworkX format that can be recognized by library gedlib.
#	Gn_new = []
#	for G in Gn:
#		G_new = nx.Graph()
#		for nd, attrs in G.nodes(data=True):
#			G_new.add_node(str(nd), chem=attrs['atom'])
#		for nd1, nd2, attrs in G.edges(data=True):
#			G_new.add_edge(str(nd1), str(nd2), valence=attrs['bond_type'])
##			G_new.add_edge(str(nd1), str(nd2))
#		Gn_new.append(G_new)
#	print(Gn_new[1].nodes(data=True))
#	print(Gn_new[1].edges(data=True))
#	print(Gn_new[1])
#	filename = '/media/ljia/DATA/research-repo/codes/others/gedlib/tests_linlin/generated_datsets/monoterpenoides/gxl/monoterpenoides'
#	xparams = {'method': 'gedlib'}
#	saveDataset(Gn, y, gformat='gxl', group='xml', filename=filename, xparams=xparams)

	# save dataset.
#	ds = {'name': 'MUTAG', 'dataset': '../../datasets/MUTAG/MUTAG.mat',
#		  'extra_params': {'am_sp_al_nl_el': [0, 0, 3, 1, 2]}}  # node/edge symb
#	Gn, y = loadDataset(ds['dataset'], extra_params=ds['extra_params'])
#	saveDataset(Gn, y, group='xml', filename='temp/temp')

	# test - new way to add labels and attributes.
#	dataset = '../../datasets/SYNTHETICnew/SYNTHETICnew_A.txt'
# 	filename = '../../datasets/Fingerprint/Fingerprint_A.txt'
#	dataset = '../../datasets/Letter-med/Letter-med_A.txt'
#	dataset = '../../datasets/AIDS/AIDS_A.txt'
#	dataset = '../../datasets/ENZYMES_txt/ENZYMES_A_sparse.txt'
# 	Gn, targets, label_names = load_dataset(filename)
	pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 27 19:33:51 2020

@author: ljia
"""
import os
import numpy as np
import pickle
from gklearn.dataset import Dataset
from gklearn.model_learning import NestedCV
from gklearn.kernels import GRAPH_KERNELS

class Workflow(object):


	def __init__(self, **kwargs):
		self._job_prefix = kwargs.get('job_prefix', 'gktask')
		self._max_num_running_tasks = kwargs.get('max_num_running_tasks', np.inf)
		self._root_dir = kwargs.get('root_dir', 'outputs/')


	def run(self, tasks):
		### Check inputs.
		if self._check_inputs(tasks):
			self._tasks = tasks
		else:
			raise ValueError('The input "tasks" is not correct.')


		### Sort tasks.
		self.sort_tasks_by_complexity()


		### The main process.
		complete = False
		while not complete:

			self.get_running_tasks()

			if self._num_running_tasks < self._max_num_running_tasks:

				### Load results from table.
				self.load_results_from_table()

				for task in self._tasks:
					state = self.get_task_state(task)
					if state != 'complete' and state != 'runnning':
						self.run_task(task)

					if self._num_running_tasks >= self._max_num_running_tasks:
						break

			### Save results.
			self.save_results()

			complete = self.check_completeness()

# 			sleep()


	def _check_inputs(self, tasks):
		if not isinstance(tasks, list):
			return False
		else:
			for i in tasks:
				if not 'kernel' in i or not 'dataset' in i:
					return False
		return True


	def sort_tasks_by_complexity(self):
		return


	def get_running_tasks(self):
		command = 'squeue --user $USER --format "%.50j" --noheader'
		stream = os.popen(command)
		output = stream.readlines()
		running_tasks = [o for o in output if o.strip().startswith(self._job_prefix)]
		self._num_running_tasks = len(running_tasks)


	def load_results_from_table(self):
		pass


	def get_task_state(self, task):
		task_dir = os.path.join(self._root_dir, task['kernel'] + '.' + task['dataset'] + '/')
		fn_summary = os.path.join(task_dir, 'results_summary.pkl')
		if os.path.isfile(fn_summary):
			output = pickle.loads(fn_summary)
			state = output['state']
			return state
		else:
			return 'unstarted'


	def run_task(self, task):
		ds_name = task['dataset']
		k_name = task['kernel']

		# Get dataset.
		ds = Dataset(ds_name)
		graph_kernel = GRAPH_KERNELS[k_name]

		# Start CV.
		results = NestedCV(ds, graph_kernel)